import pandas as pd 
import logging
import random 
import numpy as np
import multiprocessing
import itertools
import argparse
from collections import defaultdict
import pickle
import os
from os import path
from datetime import datetime
from kstar import helpers, config
import random
import string
import warnings
import math
#add import required to generate unique hash for the run_information.txt
import hashlib

#%%
#%%

class Pruner:
    """
    Pruning Algorithm used for KSTAR. 
        
    Parameters
    -----------
    network : pandas df
        weighted kinase-site prediction network where there is an accession, site, kinase, and score column
    logger : 
        logger used for pruning
    phospho_type : str
        phospho_type(s) to use when building pruned networks
    acc_col : str
        the name of the column containing Uniprot Accession IDs for each substrate in the weighted network
    site_col : str
        the name of the column containing the residue type and location of each substrate in the weighted network (Y1268, S44, etc.)
    nonweight_cols : list
        indicates the non-weight containing columns in the network (these will be removed in the final processed network, as they are not needed). If None, will automatically look
            for any non-numeric columns and removes them.
    """
    def __init__(self, network, logger, phospho_type = 'Y',acc_col = 'substrate_acc', site_col = 'site', nonweight_cols = ['substrate_acc','site','substrate_id', 'substrate_name', 'pep']):
        

        self.phospho_type = phospho_type
        self.logger = logger
        self.network = network[network[site_col].str.startswith(tuple(self.phospho_type))]
        
        #load human reference phosphoproteome, obtain the correct phosphosites (Y or ST), and combine duplicate entries
        self.compendia = config.HUMAN_REF_COMPENDIA
        self.compendia = self.compendia[self.compendia[config.KSTAR_SITE].str.startswith(tuple(self.phospho_type))]
        self.compendia = self.compendia[[config.KSTAR_ACCESSION, config.KSTAR_SITE, 'KSTAR_NUM_COMPENDIA']]
        self.compendia = self.compendia.groupby([config.KSTAR_ACCESSION, config.KSTAR_SITE]).max().reset_index()
        #merge the human phosphoproteome compendia, which contains study bias information for all phosphosites, with the network
        self.network = pd.merge(self.network, self.compendia, how = 'inner', left_on = [acc_col, site_col], right_on=[config.KSTAR_ACCESSION, config.KSTAR_SITE] )
        
        print(f'There are {self.network.shape[0]} unique substrates found in the weighted network that mapped to KinPred reference phosphoproteome\n')
        
        
        #update index to indicate site associated with each row
        self.network = self.network.set_index([config.KSTAR_ACCESSION, config.KSTAR_SITE])
        #drop unneeded columns (or maybe drop any non number columns)
        if nonweight_cols is not None:
            nonweight_cols = [col for col in nonweight_cols if col in self.network.columns]
            self.network = self.network.drop(columns=nonweight_cols)
            
        #check to make sure all remaining columns contain only numeric values
        non_numeric_cols = self.network.select_dtypes(exclude = np.number).columns
        if len(non_numeric_cols) > 0:
            print('The following non-numeric columns were identified and removed:', ', '.join(non_numeric_cols))
            print('It is assumed that these columns do not contain kinase-substrate weights. If this is incorrect, please make sure weight containing columns only contain numeric values or indicate non-weight containing columns via "nonweight_cols" parameter\n')
            self.network = self.network.select_dtypes(include = np.number)

        self.network = self.network.replace('-',np.nan)
        self.network = self.network.dropna(axis='columns', how ='all')
        
        #check for duplicate entries/indices
        if len(self.network.index.duplicated()) > 0:
            tmp = self.network[self.network.index.duplicated(keep = False)]
            #check to see if duplicate entries contain different information. If they do not, drop one row. If they do, return error.
            if any([i and v for i,v in zip(self.network.index.duplicated(keep = False), ~self.network.duplicated(keep = False).values)]):
                raise ValueError('Duplicate entries found in network that contain conflicting weights, please fix before running pruning')

            else:
                print('Duplicate entries found in network, removing repeat rows\n')
                self.network = self.network[~self.network.index.duplicated()]

        self.kinases = list(self.network.columns)
        self.kinases.remove('KSTAR_NUM_COMPENDIA')
        print(f'There are {len(self.kinases)} {phospho_type} kinases found in the weighted network')
        print(f'There are {self.network.shape[0]} unique substrates found in the weighted network that mapped to KinPred reference phosphoproteome\n')

        self.logger.info("Pruning initialization complete")
    
    def build_pruned_network(self, network, kinase_size, site_limit):
        """
        Builds a heuristic pruned network where each kinase has a specified number of connected sites and
        each site has an upper limit to the number of kinases it can connect to

        Parameters
        ----------
        network : pandas DataFrame
            network to build pruned network on
        kinase_size: int
            number of sites each kinase should connect to
        site_limit :int
            upper limit of number of kinases a site can connect to
        
        Returns
        ---------
        pruned network : pandas DataFrame
            subset of network that has been pruned
        """
        network = network.copy()
        
        site_sizes = defaultdict()
        for i in network.index:
            site_sizes[i] = 0
        num_edges = {}
        pruned_network = pd.DataFrame(index = network.index, columns = network.columns, data=np.nan)
        for i in range(kinase_size):
            random.shuffle(self.kinases)
            for kinase in self.kinases:
                if network[kinase].isna().all():
                    if kinase not in num_edges:
                        num_edges[kinase] = i+1
                        logger.info(f'{kinase} has no more available edges. Total edge count for this network = {num_edges[kinase]}')
                else:
                    sample = network.sample(weights=kinase).iloc[0]
                    network.at[sample.name, kinase] = np.nan
                    pruned_network.at[sample.name, kinase] = 1
                    site_sizes[sample.name] = site_sizes[sample.name] + 1
                    if site_sizes[sample.name] >= site_limit:
                        network.loc[sample.name,:] = np.nan 


        
        pruned_network = pruned_network.reset_index().melt(id_vars=[config.KSTAR_ACCESSION, config.KSTAR_SITE]).dropna()
        pruned_network = pruned_network.rename(columns={'variable':config.KSTAR_KINASE})
        pruned_network = pruned_network.drop(columns=['value'])   
        return pruned_network
        
    
    def compendia_pruned_network(self, compendia_sizes, site_limit, odir):
        """
        Builds a compendia-pruned network that takes into account compendia size limits per kinase
        
        Parameters
        ----------
        compendia_sizes : dict
            key : compendia size
            value : number of sites to connect to kinase 
        site_limit : int 
            upper limit of number of kinases a site can connect to
        
        Returns
        -------
        pruned_network : pandas DataFrame
            subset of network that has been pruned according to compendia ratios
        """
        pruned_network = []
        name = os.path.join(odir,''.join(random.choices(string.ascii_uppercase + string.digits, k=10)))
        for comp_size, kinase_size in compendia_sizes.items():
            self.logger.info(f"Building pruned network for compendia {comp_size} with {kinase_size} sites per kinase")
            network = self.network[self.network['KSTAR_NUM_COMPENDIA'] == comp_size]
            pruned_network.append(self.build_pruned_network(network, kinase_size, site_limit))
        pruned_network = pd.concat(pruned_network)

        pruned_network.to_csv(f"{name}.tsv", sep = "\t", index=False)

        
    def calculate_compendia_sizes(self, kinase_size):
        """
        Calculates the number of sites per compendia size that a kinase should connect to using same ratios 
        of compendia sizes as found in compendia

        Parameters
        ----------
        kinase_size: int
            number of sites each kinase should connect to
            
        Returns
        -------
        sizes : dict
            key : compendia size
            value : number of sites each kinase should pull from given compendia size
        """
        self.logger.info(f"Calculating compendia kinase sizes when kinase size is {kinase_size}")
        sizes = [-np.inf]
        multiplier = kinase_size
        while sum(sizes) < kinase_size:
            sizes = self.compendia.groupby('KSTAR_NUM_COMPENDIA').size() / len(self.compendia) * multiplier
            sizes = sizes.astype(int)
            multiplier = multiplier + 1
        sizes = sizes.to_dict()
        self.logger.info(f"Compendia kinase sizes : {sizes}")
        return sizes

    def build_multiple_compendia_networks(self, kinase_size, site_limit, num_networks, network_id, odir, PROCESSES = 1):
        """
        Builds multiple compendia-limited networks
        
        Parameters
        ----------
        
        kinase_size: int
            number of sites each kinase should connect to
        site_limit :int
            upper limit of number of kinases a site can connect to
        
        num_networks: int
            number of networks to build
        network_id : str
            id to use for each network in dictionary
        
        Returns
        -------
        pruned_networks : dict
            key : <network_id>_<i>
            value : pruned network
        """
        self.kinase_size = kinase_size
        self.site_limit = site_limit
        self.num_networks = num_networks
        self.network_id = network_id
        self.network_dir = odir
        self.use_compendia = True
        
        compendia_sizes = self.calculate_compendia_sizes(kinase_size)
        
        odir = os.path.join(odir, "work")
        if not path.exists(odir):
            os.mkdir(odir)
        num_existing_networks = len(os.listdir(odir))
        num_networks = num_networks - num_existing_networks
        if num_existing_networks > 0:
            self.logger.info(f"Found {num_existing_networks} networks... building {num_networks} additional networks")
        # MULTIPROCESSING
        if PROCESSES > 1:
            pool = multiprocessing.Pool(processes = PROCESSES)
            comp_sizes = itertools.repeat(compendia_sizes, num_networks)
            limits = itertools.repeat(site_limit, num_networks)
            odirs = itertools.repeat(odir, num_networks)
            iterable = zip(comp_sizes, limits, odirs)

            pool.starmap(self.compendia_pruned_network, iterable)
        
        # SINGLE CORE PROCESSING
        else:
            pruned_networks = []
            for i in range(num_networks):
                self.compendia_pruned_network(compendia_sizes, site_limit, odir )
        
  

    def build_multiple_networks(self, kinase_size, site_limit, num_networks, network_id, odir, PROCESSES = 1):
        """
        Basic Network Generation - only takes into account score when determining sites a kinase 
        connects to
        """
        self.kinase_size = kinase_size
        self.site_limit = site_limit
        self.num_networks = num_networks
        self.network_id = network_id
        self.network_dir = odir
        self.use_compendia = False
        
        odir = os.path.join(odir, "work")
        if not path.exists(odir):
            os.mkdir(odir)
        # MULTIPROCESSING
        if PROCESSES > 1:
            pool = multiprocessing.Pool(processes = PROCESSES)
            network_iter = itertools.repeat(self.network, num_networks)
            size_iter = itertools.repeat(kinase_size, num_networks)
            limit_iter = itertools.repeat(site_limit, num_networks)
            iterable = zip(network_iter, size_iter, limit_iter)
            pruned_networks = pool.starmap(self.build_pruned_network, iterable)
            #save each network
            for i in range(len(pruned_networks)):
                net = pruned_networks[i]
                name = os.path.join(odir,''.join(random.choices(string.ascii_uppercase + string.digits, k=10)))
                net.to_csv(f"{name}.tsv", sep = "\t", index=False) 
        # SINGLE CORE PROCESSING
        else:
            pruned_networks = []
            for i in range(num_networks):
                name = os.path.join(odir,''.join(random.choices(string.ascii_uppercase + string.digits, k=10)))
                net = self.build_pruned_network(self.network, kinase_size, site_limit)
                net.to_csv(f"{name}.tsv", sep = "\t", index=False)
                
                
    def getMaximumKinaseSize(self, site_limit):
        """
        Given a network and site_limit (maximum number of kinases a phosphorylation site can provide evidence to), will calculate the theoretical maximum number of connections each kinase can have (kinase_size parameter)
        
        Theoretical maximum exists when each substrate hits the maximum site_limit
        
        Parameters
        ----------
        site_limit: int
            Parameter used in pruning: indicates the maximum number of kinases a phosphorylation site can be connected to in the final pruned network
        
        Returns
        -------
        theoretical_max_ksize: int
            largest possible value that 'kinase_size' parameter can have without throwing any errors
        """
        
        num_substrates = self.network.shape[0]
        
        theoretical_max_ksize = (num_substrates*site_limit)/len(self.kinases)
        return theoretical_max_ksize
        
        
    def getRecommendedKinaseSize(self, site_limit):
        """
        Given a network and site_limit (maximum number of kinases a phosphorylation site can provide evidence to), will calculate the theoretical maximum number of connections each kinase can have (kinase_size parameter) and recommend a range of values for kinase_size
        
        Theoretical maximum exists when each substrate hits the maximum site_limit
        
        Parameters
        ----------
        site_limit: int
            Parameter used in pruning: indicates the maximum number of kinases a phosphorylation site can be connected to in the final pruned network
        
        Returns
        -------
        Nothing, prints theoretical maximum of kinase size and the recommened values for the parameter given the site_limit
        """
        #get the total number of substrates in the network
        theoretical_max_ksize = self.getMaximumKinaseSize(site_limit)
        print(f'Theoretical maximum value of kinase_size parameter is {theoretical_max_ksize}, given site_limit of {site_limit}')
        print(f'Recommended value of kinase size parameter is between 20-40% of theoretical maximum: {theoretical_max_ksize*0.2} - {theoretical_max_ksize*0.4}')
        
    def checkParameters(self, kinase_size, site_limit):
        """
        Given the site_limit and kinase_size parameters to be used during pruning, raise errors if not feasible, and raise warnings if value is higher than we would recommend (>40% of the maximum kinase_size value)
        
        Parameters
        ----------
        kinase_size: int
            Parameter used in pruning: indicates the number of substrates each kinase will be connected to
        site_limit: int
            Parameter used in pruning: indicates the maximum number of kinases a phosphorylation site can be connected to in the final pruned network
            
        Returns
        -------
        Nothing, will only raise errors/warnings if parameters are not feasible
        """
        
        theoretical_max_ksize = self.getMaximumKinaseSize(site_limit)
        
        if kinase_size > theoretical_max_ksize:
            raise ValueError(f"Value of 'kinase_size' is not feasible. Please decrease 'kinase_size' to {math.floor(theoretical_max_ksize)} or increase the value of site_limit")
        elif kinase_size > theoretical_max_ksize*0.4:
            warnings.warn("Value of 'kinase_size' may be higher than desired. Check the log file for potential errors after completion")
            
        
                
    def save_networks(self, network_file):
        """
        Save the pruned networks generated by the 'build_multiple_networks' or 'build_multiple_compendia_networks' as a pickle to be loaded by KSTAR
        
        """
        self.logger.info("Saving pruning results")
        self.network_file_used = network_file
        print(network_file)
        if self.use_compendia:
            suffix = f"{self.phospho_type}_compendia_{self.kinase_size}_limit_{self.site_limit}"
        else:
            suffix = f"{self.phospho_type}_{self.kinase_size}_limit_{self.site_limit}"

        # TODO: rename work directory and rename files
        os.rename(os.path.join(self.network_dir, "work"), os.path.join(self.network_dir, "INDIVIDUAL_NETWORKS"))
        odir_ind = os.path.join(self.network_dir, "INDIVIDUAL_NETWORKS")
        # rename files and make network map
        temp_filenames = os.listdir(odir_ind)
        network_map = {}
        for i in range(len(temp_filenames)):
            network_map[f"{self.network_id}_{i}"] = pd.read_table(os.path.join(odir_ind,temp_filenames[i]))
            os.rename(os.path.join(odir_ind, temp_filenames[i]), os.path.join(odir_ind, f"{self.network_id}_{i}_{suffix}.tsv"))

        pickle.dump(network_map, open( f"{self.network_dir}/{self.network_id}_{suffix}.p", "wb" ))

        # Generate a unique hash
        self.unique_id = hashlib.sha256(f"{self.network_id}_{datetime.now()}".encode()).hexdigest()

        #Call save_run_information method with the unique hash
        self.save_run_information()
        
        

        
    def save_run_information(self):
        """
        Save information about the generation of networks during run_pruning, 
        including the parameters used for generation. Primarily used when running bash script.
        
        """
        with open(f"{self.network_dir}/RUN_INFORMATION.txt", "w") as info_file:
            info_file.write("*************************************************\n")
            info_file.write(f"Pruning Information for {self.network_id}\n")
            info_file.write("*************************************************\n")
            info_file.write(f"Unique ID: {self.unique_id}\n")
            info_file.write(f"Date Run\t\t{datetime.now()}\n")
            info_file.write(f"Network Used\t{self.network_file_used}\n")
            info_file.write(f"Phospho Type\t{self.phospho_type}\n")
            info_file.write(f"Kinase Size\t\t{self.kinase_size}\n")
            info_file.write(f"Site Limit\t\t{self.site_limit}\n")
            info_file.write(f"# of Networks\t{self.num_networks}\n")
            info_file.write(f"Use Compendia\t{self.use_compendia}\n")
            if self.use_compendia:
                compendia_sizes = self.calculate_compendia_sizes(self.kinase_size)
                for comp, size in compendia_sizes.items():
                    info_file.write(f"\tCompendia {comp}\t{size}\n")


"""
********************** RUNNING PRUNING ALGORITHM FROM COMMAND LINE ****************************
Arguments
---------
--network_file          Experiment file location. csv or tsv file           (required)
--output_directory      output directory for results                        (required)
--phospho_type          phospho type (Y, ST, ...)
--kinase_size           number of sites a kinase connects to (required)
--site_limit            upper limit of number of kinases can connect to     (required)
--num_networks          number of networks to generate                      (required)
--network_id            name of network to use in building dictionary
--use_compendia         whether to use compendia ratios to build network  



"""
def parse_args():
    parser = argparse.ArgumentParser(description='Parse Network Pruning Arguments')
    parser.add_argument('--network_file', action='store', dest= 'network_file', help='Experiment file location. csv or tsv file', required=True)
    parser.add_argument('--odir','--output_directory', action = 'store', dest='odir', help = 'output directory for results', required=True)
    parser.add_argument('--phospho_type', action='store', dest = 'phospho_type', help = 'phospho type', default='Y', choices = ['Y','ST'], required = False)
    parser.add_argument('--kinase_size', action='store', dest = 'kinase_size', type = int,help='number of sites a kinase connects to',required=True)
    parser.add_argument('--site_limit', action='store', dest='site_limit', type=int,help='upper limit of number of kinases can connect to',required=True)
    parser.add_argument('--num_networks', action='store', dest='num_networks', help='number of networks to generate',type=int, required=True)
    parser.add_argument('--network_id', action='store', dest='network_id',help='name of network to use in building dictionary', default='network', required=False)
    parser.add_argument('--use_compendia', action='store', dest='use_compendia', help = 'whether to use compendia ratios to build netwokr', default = 'yes')
    parser.add_argument('--PROCESSES', action='store', dest='PROCESSES', help = 'number of processes to run in parallel', default = 1)
    results = parser.parse_args()
    return results

def process_args(results):
    # get logger
    if results.odir is None or not (path.exists(results.odir) and path.isdir(results.odir)):
        log = logging.getLogger(results.network_id)
    else:
        log = helpers.get_logger(results.network_id, f"{results.odir}/{results.network_id}_pruning.log")
    
    #check if output directory exists
    if not (path.exists(results.odir) and path.isdir(results.odir)):
        log.error("Please provide a valid output directory")
        exit()
    # check if experiment file exists and is either csv or tsv file. 
    # Load experiment if valid
    if path.exists(results.network_file) and path.isfile(results.network_file):
        log.info("Loading Network File")
        filetype = results.network_file.split('.')[-1]
        if filetype == 'csv':
            network = pd.read_csv(results.network_file)
        elif filetype == 'tsv':
            network = pd.read_csv(results.network_file, sep = '\t')
        else:
            log.error("Unrecognized network filetype. Please use a csv or tsv file")
            exit()
    else:
        log.error("Please provide a valid network file")
        exit()


    use_compendia = helpers.string_to_boolean(results.use_compendia)
    return network, log, use_compendia


def run_pruning(network, log, phospho_type, kinase_size, site_limit, num_networks, network_id, odir, use_compendia = True,
                acc_col = 'substrate_acc', site_col = 'site', netcols_todrop = ['substrate_acc','site','substrate_id', 'substrate_name', 'pep'], PROCESSES = 1):
    """
    Generate pruned networks from a weighted kinase-substrate graph and log run information
    
    Parameters
    ----------
    network: pandas dataframe
        kinase substrate network matrix, with values indicating weight of kinase-substrate relationship
    log: logger
        logger to document the pruning process from start to finish
    use_compendia: string
        whether to use compendia ratios to build network
    phospho_type: string
        phospho type ('Y', 'ST', ...)
    kinase_size: int
        number of sites a kinase connects to
    site_limit: int
        upper limit of number of kinases can connect to
    num_networks: int
        number of networks to generate
    network_id: string
        name of network to use in building dictionary
    odir: string
        output directory for results
        
    Returns
    -------
    pruner: Prune object
        prune object that contains the number of pruned networks indicated by the num_networks paramater
    """
    log.info("Running pruning algorithm")
    pruner = Pruner(network, log, phospho_type, acc_col = acc_col, site_col = site_col, nonweight_cols = netcols_todrop)
    pruner.checkParameters(kinase_size, site_limit)
    if use_compendia:
        log.info("Pruning using compendia ratios")
        pruner.build_multiple_compendia_networks(kinase_size, site_limit, num_networks, network_id, odir, PROCESSES = PROCESSES)
    else:
        log.info("Pruning without using compendia")
        pruned_networks = pruner.build_multiple_networks(kinase_size, site_limit, num_networks, network_id, odir, PROCESSES = PROCESSES)
    return pruner

def save_pruning(phospho_type, network_id, kinase_size, site_limit, use_compendia, odir, log):
    """
    Save the pruned networks generated by run_pruning function as a pickle to be loaded by KSTAR
    
    Parameters
    ----------
    phosho_type: string
        type of phosphomodification to networks were generated for (either 'Y' or 'ST')
    network_id: string
        name of network used to build dictionary
    kinase_size: int
        number of sites a kinase connects to
    site_limit: int
        upper limit of number of kinases can connect to
    use_compendia: string 
        whether compendia was used for ratios to build networks
    odir: string
        output directory for results
    log: logger
        logger to document pruning process from start to finish
        
    Returns
    -------
    Nothing
    """
    log.info("Saving pruning results")
    # if not os.path.exists(f"{odir}/INDIVIDUAL_NETWORKS"): 
    #         os.mkdir(f"{odir}/INDIVIDUAL_NETWORKS") 
    
    if use_compendia:
        suffix = f"{phospho_type}_compendia_{kinase_size}_limit_{site_limit}"
    else:
        suffix = f"{phospho_type}_{kinase_size}_limit_{site_limit}"

    # TODO: rename work directory and rename files
    os.rename(os.path.join(odir, "work"), os.path.join(odir, "INDIVIDUAL_NETWORKS"))
    odir_ind = os.path.join(odir, "INDIVIDUAL_NETWORKS")
    # rename files and make network map
    temp_filenames = os.listdir(odir_ind)
    network_map = {}
    for i in range(len(temp_filenames)):
        network_map[f"{network_id}_{i}"] = pd.read_table(os.path.join(odir_ind,temp_filenames[i]))
        os.rename(os.path.join(odir_ind, temp_filenames[i]), os.path.join(odir_ind, f"{network_id}_{i}_{suffix}.tsv"))

    # for nid, network in network_map.items():
    #     network.to_csv(f"{odir}/INDIVIDUAL_NETWORKS/{nid}_{suffix}.tsv", sep = '\t', index=False)
    pickle.dump( network_map, open( f"{odir}/{network_id}_{suffix}.p", "wb" ))
    # Generate a unique hash
    unique_id = hashlib.sha256(f"{network_id}_{datetime.now()}".encode()).hexdigest()


    
def save_run_information(results, use_compendia, pruner, unique_id):
    """
    Save information about the generation of networks during run_pruning, 
    including the parameters used for generation. Primarily used when running bash script.
    
    Parameters
    ----------
    results: 
        object that stores all parameters used in the pruning process
    use_compendia: string
        whether compendia was used for ratios to build network
    pruner: Prune object
        output of the run_pruning() function
        
    Returns
    -------
    Nothing
    """
    with open(f"{results.odir}/RUN_INFORMATION.txt", "w") as info_file:
        info_file.write("*************************************************\n")
        info_file.write(f"Pruning Information for {results.network_id}\n")
        info_file.write("*************************************************\n")
        info_file.write(f"Unique ID:{unique_id}\n")
        info_file.write(f"Date Run\t\t{datetime.now()}\n")
        info_file.write(f"Network Used\t{results.network_file}\n")
        info_file.write(f"Phospho Type\t{results.phospho_type}\n")
        info_file.write(f"Kinase Size\t\t{results.kinase_size}\n")
        info_file.write(f"Site Limit\t\t{results.site_limit}\n")
        info_file.write(f"# of Networks\t{results.num_networks}\n")
        info_file.write(f"Use Compendia\t{results.use_compendia}\n")
        if use_compendia:
            compendia_sizes = pruner.calculate_compendia_sizes(results.kinase_size)
            for comp, size in compendia_sizes.items():
                info_file.write(f"\tCompendia {comp}\t{size}\n")



def main():
    results = parse_args()
    network, log, use_compendia = process_args(results)
    phospho_type = results.phospho_type
    kinase_size = results.kinase_size
    site_limit = results.site_limit
    num_networks = results.num_networks
    network_id = results.network_id
    odir = results.odir
    try:
        PROCESSES = int(results.PROCESSES)
    except:
        log.info('PROCESSES parameter must be an integer')
        exit()
        
    log.info("Beginning to build pruning networks")
    pruner = run_pruning(network, log, phospho_type, kinase_size, site_limit, num_networks, network_id, odir)
    save_pruning(phospho_type, network_id, kinase_size, site_limit, use_compendia, odir, log)
    save_run_information(results, use_compendia, pruner)

if __name__ == "__main__":
    main()






import pandas as pd 
import numpy as np
from collections import defaultdict


def getSubstrateInfluence(networks, kinase, substrate_subset = None):
    """
    Given the pruned networks and kinase of interest, return the number of networks each substrate is connected to that kinase in (the 'substrate influence' on that kinase's activity prediction). If subset of substrates is provided, will only do this for the given subset
    
    Parameters
    ----------
    networks: dict
        dictionary storing all pruned networks used in activity calculation
    kinase: str
        name of the kinase of interest: should match the name found in provided networks
    substrate_subset: list
        subset of substrates to analyze, indicated by '{KSTAR_ACCESSION}_{KSTAR_SITE}'. If none, will return a series containing info on all substrates with at least one connection to the given kinase
    
    Returns
    -------
    Pandas series indicating the number of networks each substrate is connected the indicated kinase, sorted from the most connections (highest influence) to the least (lowest influence). Sites with no connection will not be included.
    """
    num_networks = defaultdict(int)
    for nname in networks:
        net = networks[nname]
        net_trim = net[net['KSTAR_KINASE'] == kinase].copy()
        net_trim['KSTAR_SUBSTRATE'] = net_trim['KSTAR_ACCESSION']+'_'+net_trim['KSTAR_SITE']
        
        #if subset of substrates provided, reduce to this size
        if substrate_subset is not None:
            keep = [True if sub in substrate_subset else False for sub in net_trim['KSTAR_SUBSTRATE'].values]
            net_trim = net_trim[keep]
        
        #loop through each substrate in network and add to num_network dict if present
        for sub in net_trim['KSTAR_SUBSTRATE']:
            num_networks[sub] += 1
        
    return pd.Series(num_networks, dtype = float, name = f'Number of Networks connected to {kinase}').sort_values(ascending = False)
    
def getSubstrateInfluence_inExperiment(networks, binary_evidence, kinase, data_cols = None):
    """
    Given the binary evidence used for activity prediction, identify which sites are found across the most networks for a given kinase and each sample. 
    
    Parameters
    ----------
    networks: dictionary
        dictionary containing all 50 pruned networks used for activity prediction
    binary_evidence: pandas dataframe
        binarized dataset (using the same threshold/criteria as the one used for activity prediction)
    kinase: str
        name of the kinase to probe
    data_cols: list or None
        name of the data columns in binary_evidence to probe. If None, will analyze all columns with 'data:' at the start of the column name.
    """
    experiment = binary_evidence.copy()
    if data_cols is None:
        data_cols = [col for col in experiment.columns if 'data:' in col]
    
    experiment['KSTAR_SUBSTRATE'] = experiment['KSTAR_ACCESSION'] + '_' + experiment['KSTAR_SITE']
    experiment_influence = {}
    for col in data_cols:
        substrate_subset =  experiment.loc[experiment[col] == 1,'KSTAR_SUBSTRATE'].values
        experiment_influence[col] = getSubstrateInfluence(networks, kinase, substrate_subset = substrate_subset)
    return experiment_influence
    
    
    
"""
def plotMannWhitney(real_enrichment, random_enrichment, kinase, data_col, mann_whitney_activities = None, mann_whitney_fpr = None):
    #transform enrichment values and plot
    transformed_real = -np.log10(real_enrichment.loc[(real_enrichment['kinase'] == kinase) & (real_enrichment['data'] == data_col), 'kinase_activity'].values)
    plt.hist(transformed_real, alpha = 0.5, bins = 50, label = 'Real')
    transformed_random = -np.log10(random_enrichment['kinase'] == kinase) & (random_enrichment['data'] == data_col) 'kinase_activity'].values)
    plt.hist(transformed_random, alpha = 0.5, bins = 50, label = 'Random')
    plt.legend()
    
    #annotate plot with mann whitney results if provided
    activity = -np.log10(mann_whitney_activities.loc[kinase, data_col])
    plt.annotate(
"""


import pickle
import os
import pandas as pd
from kstar import helpers
import requests
import tarfile

"""
These are the GLOBAL variables for the KSTAR project and are set for default procedures for the current release
of KinPred and KSTAR. 

Update these if you want to update:
1. Where resources are placed from KSTAR and KINPRED projects: RESOURCE_DIR
2. Reference proteome: HUMAN_REF_FASTA_FILE
3. Phosphoprotoeme: HUMAN_REF_PHOSPHO
4. Networks Directory: Downloaded from KSTAR Figshare repository, this includes heuristic networks suggested for 
5. Networks Pickle: 

"""

## BEGIN DECLARATION OF GLOBALS

KSTAR_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
RESOURCE_DIR = KSTAR_DIR + "/RESOURCE_FILES"

# DOWNLOAD RESOURCE_FILES from Figshare to get started using networks
# Directories for the project (where to find resource file folder), resource folder, and network directories are all set by directories.txt, which
# can be updated with update_directories().
directory_file = open(f'{KSTAR_DIR}/kstar/directories.txt', 'r')
directories = []
for line in directory_file:
    directories.append(line.split()[0])
directory_file.close()

NETWORK_DIR = directories[0]

RESOURCE_URL = 'https://ndownloader.figshare.com/files/28762653'  # location of corresponding release of data
NETWORK_URL = 'https://ndownloader.figshare.com/files/52178324'

# RESOURCE_DIR dependencies
HUMAN_REF_FASTA_FILE = f"{RESOURCE_DIR}/humanProteome.fasta"  # download from KinPred https://doi.org/10.1101/2020.08.10.244426
try:
    HUMAN_REF_SEQUENCES = helpers.process_fasta_file(HUMAN_REF_FASTA_FILE)
except FileNotFoundError:
    HUMAN_REF_SEQUENCES = None
    print(
        'Could not find reference proteome. Please install the resource directory using config.install_resource_files()')

HUMAN_REF_PHOSPHO_FILE = f"{RESOURCE_DIR}/HumanPhosphoProteome.csv"  # download from KSTAR FIGSHARE, or use helpers folder generate to create a new one
try:
    HUMAN_REF_COMPENDIA = pd.read_csv(HUMAN_REF_PHOSPHO_FILE)
except FileNotFoundError:
    HUMAN_REF_COMPENDIA = None
    print(
        'Could not find reference proteome. Please install the resource directory using config.install_resource_files()')

NETWORK_Y_PICKLE = f"{NETWORK_DIR}/network_Y.p"  # created by create_networkin_pickles()
NETWORK_ST_PICKLE = f"{NETWORK_DIR}/network_ST.p"  # created by create_networkin_pickles()

# check to see if networks can be found
if not os.path.isdir(NETWORK_DIR):
    print('Could not find network directory. Please update using config.update_network_directory().')
elif not os.path.isfile(NETWORK_Y_PICKLE) and not os.path.isfile(NETWORK_ST_PICKLE):
    print('Could not find any network pickles. Please create network pickles using config.create_network_pickles().')

# COLUMN NAMES USED FOR KSTAR
KSTAR_ACCESSION = 'KSTAR_ACCESSION'
KSTAR_PEPTIDE = 'KSTAR_PEPTIDE'
KSTAR_SITE = 'KSTAR_SITE'
KSTAR_KINASE = 'KSTAR_KINASE'

# PARAMETERS USED FOR RUNNING KSTAR USING PREGENERATED EXPERIMENTS
USE_PREGEN_DATA = True
SAVE_NEW_PRECOMPUTE = False
PREGENERATED_EXPERIMENTS_PATH = NETWORK_DIR
NETWORK_HASH_Y = 'd73e971979661777c05182efad8b314b1aa78896e242810cce8ecd51b23c85c6'
NETWORK_HASH_ST = '68328e443909aecbb6a1f6b0572da171f6b1213c95ecf1928b621ef39826c953'

## END DECLARATION OF GLOBALS


def install_resource_files():
    """Retrieves RESOURCE_FILES that are the companion for this version release from FigShare, unzips them to the correct directory for resource files."""

    print("Requesting resource file")
    r = requests.get(RESOURCE_URL, allow_redirects=True)
    outputFile = KSTAR_DIR + "/RESOURCE_FILES.tar.gz"

    open(outputFile, 'wb').write(r.content)

    t = tarfile.open(outputFile, 'r')

    print("Extracting %s" % outputFile)
    t.extractall(KSTAR_DIR)


# def update_network_directory(directory, *kwargs):

def install_network_files(target_dir=None):
    """Retrieves Network files that are the companion for this version release from FigShare, unzips them to the specified directory."""
    global DIRECTORY_FOR_SAVE_PRECOMPUTE

    print("Requesting network file")
    r = requests.get(NETWORK_URL, allow_redirects=True)

    # Use target_dir if provided, otherwise default to KSTAR_DIR
    install_dir = target_dir if target_dir else KSTAR_DIR
    outputFile = os.path.join(install_dir, "NETWORKS.tar.gz")

    # Create target directory if it doesn't exist
    os.makedirs(install_dir, exist_ok=True)

    open(outputFile, 'wb').write(r.content)

    t = tarfile.open(outputFile, 'r')

    print(f"Extracting {outputFile}")
    t.extractall(install_dir)

def update_network_directory(directory, create_pickles=True, KSTAR_DIR=KSTAR_DIR, NETWORK_DIR=NETWORK_DIR):
    """
    Update the location of network the network files, and verify that all necessary files are located in directory

    Parameters
    ----------
    directory: string
        path to where network files are located


    """

    # check that directory exists
    if not os.path.isdir(directory):
        print(
            'Directory not found, so directories.txt was not updated and no pickles were generated. Please verify that directory is correct')
        return

    # update network directory in config (temporary change)
    NETWORK_DIR = directory
    # update network directory in directories.txt (permanent change)
    with open(f'{KSTAR_DIR}/kstar/directories.txt', 'w') as d:
        d.write(directory)
    print('Network directory updated.')

    # If create_pickles is true, look for individual networks and create pickles
    if create_pickles:
        if not os.path.isfile(f"{directory}/network_Y.p") and not os.path.isfile(f"{directory}/network_ST.p"):
            if os.path.isdir(f'{directory}/Y/INDIVIDUAL_NETWORKS') and os.path.isdir(
                    f'{directory}/ST/INDIVIDUAL_NETWORKS'):
                print('Found individual networks for both Y and ST, generating pickles for both')
                create_network_pickles(network_directory=directory)
            elif os.path.isdir(f'{directory}/Y/INDIVIDUAL_NETWORKS'):
                print('Individual networks found for Y, but not for ST. Generating Y pickle only.')
                create_network_pickles(['Y'], network_directory=directory)
            elif os.path.isdir(f'{directory}/ST/INDIVIDUAL_NETWORKS'):
                print('Individual networks found for ST, but not for Y. Generating ST pickle only.')
                create_network_pickles(['ST'], network_directory=directory)
        elif not os.path.isfile(f"{directory}/network_Y.p"):
            print('ST pickle already created. Generating Y pickle')
            if os.path.isdir(f'{directory}/Y/INDIVIDUAL_NETWORKS'):
                create_network_pickles(['Y'], network_directory=directory)
            else:
                print(
                    'Could not find individual networks for Y. Make sure networks are deposited in "{NETWORK_DIR}/Y/INDIVIDUAL_NETWORKS/", then rerun create_network_pickles()')
        elif not os.path.isfile(f'{directory}/network_ST.p'):
            print('Y pickle already created. Generating ST pickle')
            if os.path.isdir(f'{directory}/ST/INDIVIDUAL_NETWORKS'):
                create_network_pickles(['ST'], network_directory=directory)
            else:
                print(
                    'Could not find individual networks for ST. Make sure networks are deposited in "{NETWORK_DIR}/ST/INDIVIDUAL_NETWORKS/", , then rerun create_network_pickles()')
        else:
            print('Network pickles already generated')

    # update location of network pickles
    NETWORK_Y_PICKLE = f"{NETWORK_DIR}/network_Y.p"  # created by create_networkin_pickles()
    NETWORK_ST_PICKLE = f"{NETWORK_DIR}/network_ST.p"  # created by create_networkin_pickles()
    return NETWORK_DIR, NETWORK_Y_PICKLE, NETWORK_ST_PICKLE


# def create_network_pickles(phosphoType = ['Y','ST'], *kwargs)
def create_network_pickles(phosphoTypes=['Y', 'ST'], network_directory=NETWORK_DIR):
    """
    Given network files declared in globals, create pickles of the kstar object that can then be quickly loaded in analysis
    Assumes that the Network structure has two folders Y and ST under the NETWORK_DIR global variable and that
    all .csv files in those directories should be loaded into a network pickle.
    """

    for phosphoType in phosphoTypes:
        network = {}
        if not os.path.isfile(f"{network_directory}/network_{phosphoType}.p"):
            directory = f"{network_directory}/{phosphoType}/INDIVIDUAL_NETWORKS/"
            # get all csv files in that directory
            for file in os.listdir(directory):
                if file.endswith(".tsv"):
                    # get the value of the network number
                    file_noext = file.strip(".tsv").split('_')
                    key_name = 'nkin' + str(file_noext[1])
                    # print("Debug: key name is %s"%(key_name))
                    network[key_name] = pd.read_csv(f"{directory}{file}", sep='\t')
            print("Loaded %d number of networks for phosphoType %s" % (len(network), phosphoType))
            pickle.dump(network, open(f"{network_directory}/network_{phosphoType}.p", "wb"))
            print(f"Saved pickle file at {network_directory}/network_{phosphoType}.p")
        else:
            print(f"{phosphoType} network pickle already generated")


def check_configuration():
    """
    Verify that all necessary files are downloadable and findable
    """
    ready_Y = True
    ready_ST = True
    if not os.path.isfile(HUMAN_REF_FASTA_FILE) or not os.path.isfile(HUMAN_REF_PHOSPHO_FILE):
        print(
            'Could not find reference proteome and/or phosphoproteome. Please install the resource directory using config.install_resource_files()')
        ready_Y, ready_ST = False, False
    if not os.path.isdir(NETWORK_DIR):
        print('Could not find network directory. Please update using config.update_network_directory().')
        ready_Y, ready_ST = False, False
    else:
        if not os.path.isfile(NETWORK_Y_PICKLE) and not os.path.isfile(NETWORK_ST_PICKLE):
            print(
                'Could not find any network pickles. Please create network pickles using config.create_network_pickles() or change network directory to where pickles are located with config.update_network_directory().')
            ready_Y, ready_ST = False, False
        elif not os.path.isfile(NETWORK_Y_PICKLE):
            ready_Y = False
        elif not os.path.isfile(NETWORK_ST_PICKLE):
            ready_ST = False

    if ready_Y and ready_ST:
        print('You are ready to begin! You can generate predictions for both Y and ST networks.')
    elif ready_Y:
        print(
            'You are ready to generate predictions for Y networks, but not ST networks. If you want to generate ST predictions, create the ST network pickle with config.create_network_pickles(phosphoType = ["ST"])')
    elif ready_ST:
        print(
            'You are ready to generate predictions for ST networks, but not Y networks. If you want to generate Y predictions, create the Y network pickle with config.create_network_pickles(phosphoType = ["Y"])')




from setuptools import setup, find_packages


# Run setup
setup(
    name="kstar",

    version="1.0.2",
    author="Naegle Lab",
    author_email="kmn4mj@virginia.edu",
    url="https://github.com/NaegleLab/KSTAR",
    install_requires=['pandas==2.1.*', 'numpy==1.26.*', 'scipy==1.11.*', 'matplotlib==3.8.*', 'seaborn==0.13.*', 'biopython==1.78.*','requests==2.31.*'],

    license='GNU General Public License v3',
    description='KSTAR: Kinase-Substrate Transfer to Activity Relationships',
    long_description="""KSTAR is an open-source software for estimating kinase activities from phosphoproteomic data. 
    KSTAR implements statistical and graph-theoretic approaches to produce a robust activity score that increases with increasing evidence from a kinase's network.""",
    project_urls = {'Documentation': 'https://naeglelab.github.io/KSTAR/'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    include_package_data = True,
    python_requires=">=3.10",
    zip_safe = False
)
